@@ -65,6 +65,8 @@ struct _DbusTestServicePrivate {
 
 	DbusTestWatchdog * watchdog;
 	guint watchdog_source;
+
+	DbusTestServiceBus bus_type;
 };
 
 #define SERVICE_CHANGE_HANDLER  "dbus-test-service-change-handler"
@@ -124,6 +126,8 @@ dbus_test_service_init (DbusTestService *self)
 	                                                         g_object_ref(self->priv->watchdog),
 	                                                         g_object_unref);
 
+	self->priv->bus_type = DBUS_TEST_SERVICE_BUS_SESSION;
+
 	return;
 }
 
@@ -241,54 +245,81 @@ watchdog_ping (gpointer user_data)
 	return TRUE;
 }
 
-static void
-all_tasks_finished_helper (gpointer data, gpointer user_data)
+static gboolean
+all_tasks_finished_helper (G_GNUC_UNUSED DbusTestService * service, DbusTestTask * task, G_GNUC_UNUSED gpointer user_data)
 {
-	DbusTestTask * task = DBUS_TEST_TASK(data);
-	gboolean * all_finished = (gboolean *)user_data;
-
 	DbusTestTaskState state = dbus_test_task_get_state(task);
 	DbusTestTaskReturn ret  = dbus_test_task_get_return(task);
 
 	if (state != DBUS_TEST_TASK_STATE_FINISHED && (ret != DBUS_TEST_TASK_RETURN_IGNORE || dbus_test_task_get_wait_finished(task))) {
-		*all_finished = FALSE;
+		return FALSE;
 	}
 
-	return;
-}
-
-static void
-all_tasks_started_helper (gpointer data, gpointer user_data)
-{
-	DbusTestTask * task = DBUS_TEST_TASK(data);
-	gboolean * all_started = (gboolean *)user_data;
-
-	DbusTestTaskState state = dbus_test_task_get_state(task);
-
-	if (state == DBUS_TEST_TASK_STATE_INIT || state == DBUS_TEST_TASK_STATE_WAITING) {
-		*all_started = FALSE;
-	}
-
-	return;
+	return TRUE;
 }
 
 static gboolean
-all_tasks (DbusTestService * service, GFunc helper)
+all_tasks_started_helper (G_GNUC_UNUSED DbusTestService * service, DbusTestTask * task, G_GNUC_UNUSED gpointer user_data)
 {
-	gboolean breaknow = TRUE;
+	DbusTestTaskState state = dbus_test_task_get_state(task);
 
-	g_queue_foreach(&service->priv->tasks_first, helper, &breaknow);
-	if (!breaknow) {
+	if (state == DBUS_TEST_TASK_STATE_INIT || state == DBUS_TEST_TASK_STATE_WAITING) {
 		return FALSE;
 	}
 
-	g_queue_foreach(&service->priv->tasks_normal, helper, &breaknow);
-	if (!breaknow) {
+	return TRUE;
+}
+
+static gboolean
+all_tasks_bus_match (DbusTestService * service, DbusTestTask * task, G_GNUC_UNUSED gpointer user_data)
+{
+	return service->priv->bus_type == DBUS_TEST_SERVICE_BUS_BOTH ||
+		dbus_test_task_get_bus(task) == DBUS_TEST_SERVICE_BUS_BOTH ||
+		dbus_test_task_get_bus(task) == service->priv->bus_type;
+}
+
+typedef struct {
+	DbusTestService * service;
+	gboolean passing;
+	gpointer user_data;
+	gboolean (*func) (DbusTestService * service, DbusTestTask * task, gpointer data);
+} all_tasks_helper_data_t;
+
+static void
+all_tasks_helper (gpointer taskp, gpointer datap)
+{
+	all_tasks_helper_data_t * data = (all_tasks_helper_data_t *)datap;
+
+	if (!data->passing) {
+		/* Quick exit */
+		return;
+	}
+
+	data->passing = data->func(data->service, DBUS_TEST_TASK(taskp), data->user_data);
+}
+
+static gboolean
+all_tasks (DbusTestService * service, gboolean (*helper) (DbusTestService * service, DbusTestTask * task, gpointer user_data), gpointer user_data)
+{
+	all_tasks_helper_data_t data = {
+		.passing = TRUE,
+		.service = service,
+		.func = helper,
+		.user_data = user_data
+	};
+
+	g_queue_foreach(&service->priv->tasks_first, all_tasks_helper, &data);
+	if (!data.passing) {
 		return FALSE;
 	}
 
-	g_queue_foreach(&service->priv->tasks_last, helper, &breaknow);
-	if (!breaknow) {
+	g_queue_foreach(&service->priv->tasks_normal, all_tasks_helper, &data);
+	if (!data.passing) {
+		return FALSE;
+	}
+
+	g_queue_foreach(&service->priv->tasks_last, all_tasks_helper, &data);
+	if (!data.passing) {
 		return FALSE;
 	}
 
@@ -367,9 +398,23 @@ dbus_writes (GIOChannel * channel, GIOCondition condition, gpointer data)
 	if (service->priv->first_time) {
 		service->priv->first_time = FALSE;
 
-		g_setenv("DBUS_SESSION_BUS_ADDRESS", line, TRUE);
 		g_setenv("DBUS_STARTER_ADDRESS", line, TRUE);
-		g_setenv("DBUS_STARTER_BUS_TYPE", "session", TRUE);
+
+		switch (service->priv->bus_type) {
+		case DBUS_TEST_SERVICE_BUS_SESSION:
+			g_setenv("DBUS_SESSION_BUS_ADDRESS", line, TRUE);
+			g_setenv("DBUS_STARTER_BUS_TYPE", "session", TRUE);
+			break;
+		case DBUS_TEST_SERVICE_BUS_SYSTEM:
+			g_setenv("DBUS_SYSTEM_BUS_ADDRESS", line, TRUE);
+			g_setenv("DBUS_STARTER_BUS_TYPE", "system", TRUE);
+			break;
+		case DBUS_TEST_SERVICE_BUS_BOTH:
+			g_setenv("DBUS_SESSION_BUS_ADDRESS", line, TRUE);
+			g_setenv("DBUS_SYSTEM_BUS_ADDRESS", line, TRUE);
+			g_setenv("DBUS_STARTER_BUS_TYPE", "session", TRUE);
+			break;
+		}
 
 		if (service->priv->state == STATE_DAEMON_STARTING) {
 			g_main_loop_quit(service->priv->mainloop);
@@ -451,8 +496,15 @@ start_daemon (DbusTestService * service)
 	g_main_loop_run(service->priv->mainloop);
 
 	/* we should have a usable connection now, let's check */
-	gchar **tokens = g_strsplit (g_getenv ("DBUS_SESSION_BUS_ADDRESS"),
-	                             ",", 0);
+	const gchar * bus_address = NULL;
+	if (service->priv->bus_type == DBUS_TEST_SERVICE_BUS_SYSTEM) {
+		bus_address = g_getenv("DBUS_SYSTEM_BUS_ADDRESS");
+	} else {
+		bus_address = g_getenv("DBUS_SESSION_BUS_ADDRESS");
+	}
+	g_return_if_fail(bus_address != NULL);
+	gchar **tokens = g_strsplit (bus_address, ",", 0);
+
 	guint i;
 	gboolean is_valid = FALSE;
 	for (i = 0; i < g_strv_length (tokens); i++) {
@@ -478,12 +530,14 @@ void
 dbus_test_service_start_tasks (DbusTestService * service)
 {
 	g_return_if_fail(DBUS_TEST_SERVICE(service));
+	g_return_if_fail(all_tasks(service, all_tasks_bus_match, NULL));
 
 	start_daemon(service);
-	g_return_if_fail(g_getenv("DBUS_SESSION_BUS_ADDRESS") != NULL);
+	g_return_if_fail(g_getenv("DBUS_SESSION_BUS_ADDRESS") != NULL ||
+		g_getenv("DBUS_SYSTEM_BUS_ADDRESS") != NULL);
 	g_return_if_fail(service->priv->state != STATE_DAEMON_FAILED);
 
-	if (all_tasks(service, all_tasks_started_helper)) {
+	if (all_tasks(service, all_tasks_started_helper, NULL)) {
 		/* If we have all started we can mark it as such as long
 		   as we understand where we could hit this case */
 		if (service->priv->state == STATE_INIT || service->priv->state == STATE_DAEMON_STARTED) {
@@ -506,12 +560,12 @@ dbus_test_service_start_tasks (DbusTestService * service)
 	}
 	g_queue_foreach(&service->priv->tasks_last, task_starter, NULL);
 
-	if (!all_tasks(service, all_tasks_started_helper)) {
+	if (!all_tasks(service, all_tasks_started_helper, NULL)) {
 		service->priv->state = STATE_STARTING;
 		g_main_loop_run(service->priv->mainloop);
 
 		/* This should never happen, but let's be sure */
-		g_return_if_fail(all_tasks(service, all_tasks_started_helper));
+		g_return_if_fail(all_tasks(service, all_tasks_started_helper, NULL));
 	}
 
 	service->priv->state = STATE_STARTED;
@@ -519,17 +573,10 @@ dbus_test_service_start_tasks (DbusTestService * service)
 	return;
 }
 
-static void
-all_tasks_passed_helper (gpointer data, gpointer user_data)
+static gboolean
+all_tasks_passed_helper (G_GNUC_UNUSED DbusTestService * service, DbusTestTask * task, G_GNUC_UNUSED gpointer user_data)
 {
-	DbusTestTask * task = DBUS_TEST_TASK(data);
-	gboolean * all_passed = (gboolean *)user_data;
-
-	if (!dbus_test_task_passed(task)) {
-		*all_passed = FALSE;
-	}
-
-	return;
+	return dbus_test_task_passed(task);
 }
 
 static int
@@ -539,7 +586,7 @@ get_status (DbusTestService * service)
 		return -1;
 	}
 
-	if (all_tasks(service, all_tasks_passed_helper)) {
+	if (all_tasks(service, all_tasks_passed_helper, NULL)) {
 		return 0;
 	} else {
 		return -1;
@@ -554,7 +601,7 @@ dbus_test_service_run (DbusTestService * service)
 	dbus_test_service_start_tasks(service);
 	g_return_val_if_fail(service->priv->state == STATE_STARTED, get_status(service));
 
-	if (all_tasks(service, all_tasks_finished_helper)) {
+	if (all_tasks(service, all_tasks_finished_helper, NULL)) {
 		return get_status(service);
 	}
 
@@ -562,7 +609,7 @@ dbus_test_service_run (DbusTestService * service)
 	g_main_loop_run(service->priv->mainloop);
 
 	/* This should never happen, but let's be sure */
-	g_return_val_if_fail(all_tasks(service, all_tasks_finished_helper), -1);
+	g_return_val_if_fail(all_tasks(service, all_tasks_finished_helper, NULL), -1);
 	service->priv->state = STATE_FINISHED;
 
 	return get_status(service);
@@ -574,12 +621,12 @@ task_state_changed (G_GNUC_UNUSED DbusTestTask * task, G_GNUC_UNUSED DbusTestTas
 	g_return_if_fail(DBUS_TEST_IS_SERVICE(user_data));
 	DbusTestService * service = DBUS_TEST_SERVICE(user_data);
 
-	if (service->priv->state == STATE_STARTING && all_tasks(service, all_tasks_started_helper)) {
+	if (service->priv->state == STATE_STARTING && all_tasks(service, all_tasks_started_helper, NULL)) {
 		g_main_loop_quit(service->priv->mainloop);
 		return;
 	}
 
-	if (service->priv->state == STATE_RUNNING && all_tasks(service, all_tasks_finished_helper)) {
+	if (service->priv->state == STATE_RUNNING && all_tasks(service, all_tasks_finished_helper, NULL)) {
 		g_main_loop_quit(service->priv->mainloop);
 		return;
 	}
@@ -599,6 +646,10 @@ dbus_test_service_add_task_with_priority (DbusTestService * service, DbusTestTas
 	g_return_if_fail(DBUS_TEST_IS_SERVICE(service));
 	g_return_if_fail(DBUS_TEST_IS_TASK(task));
 
+	/* Check to ensure that the task and the service match in thier
+	   goals for busness. Fail early. */
+	g_return_if_fail(all_tasks_bus_match(service, task, NULL));
+
 	GQueue * queue = NULL;
 
 	switch (prio) {
@@ -624,6 +675,45 @@ dbus_test_service_add_task_with_priority (DbusTestService * service, DbusTestTas
 	return;
 }
 
+/**
+ * @service: A #DbusTestService
+ * @task: Task to remove
+ *
+ * Removes a task from those managed by the service, it won't
+ * be checked for status or managed anymore by the service.
+ *
+ * Return Value: Whether the task was found and removed, FALSE if not found
+ */
+gboolean
+dbus_test_service_remove_task (DbusTestService * service, DbusTestTask * task)
+{
+	g_return_val_if_fail(DBUS_TEST_IS_SERVICE(service), FALSE);
+	g_return_val_if_fail(DBUS_TEST_IS_TASK(task), FALSE);
+
+	guint count = 0;
+	count += g_queue_remove_all(&service->priv->tasks_first, task);
+	count += g_queue_remove_all(&service->priv->tasks_normal, task);
+	count += g_queue_remove_all(&service->priv->tasks_last, task);
+
+	/* Checking the count here so that we can generate a warning. Guessing that
+	   this actually never happens, but it's easy to check */
+	if (count > 1) {
+		g_warning("Task '%s' was added to the service %d times!", dbus_test_task_get_name(task), count);
+	}
+
+	/* We're going to disconnect here even if count is zero because, well, it
+	   shouldn't hurt in that case and might be good for us. */
+	g_signal_handlers_disconnect_by_data(task, service);
+
+	/* If we've added it multiple times, we made multiple references, fix it. */
+	guint i;
+	for (i = 0; i < count; i++) {
+		g_object_unref(task);
+	}
+
+	return count > 0;
+}
+
 void
 dbus_test_service_set_daemon (DbusTestService * service, const gchar * daemon)
 {
@@ -656,3 +746,25 @@ dbus_test_service_stop (DbusTestService * service)
 	g_main_loop_quit(service->priv->mainloop);
 	return;
 }
+
+void dbus_test_service_set_bus (DbusTestService * service, DbusTestServiceBus bus)
+{
+	g_return_if_fail(DBUS_TEST_IS_SERVICE(service));
+	g_return_if_fail(service->priv->dbus == 0); /* we can't change after we're running */
+
+	if (bus == DBUS_TEST_SERVICE_BUS_BOTH) {
+		g_warning("Setting bus to BOTH, which is typically only used as a default value.");
+	}
+
+	service->priv->bus_type = bus;
+	g_warn_if_fail(all_tasks(service, all_tasks_bus_match, NULL));
+
+	if (bus == DBUS_TEST_SERVICE_BUS_SYSTEM) {
+		g_free(service->priv->dbus_configfile);
+		service->priv->dbus_configfile = g_strdup(DEFAULT_SYSTEM_CONF);
+	} else {
+		/* BOTH and SESSION get the session config — for backwards compatibility there */
+		g_free(service->priv->dbus_configfile);
+		service->priv->dbus_configfile = g_strdup(DEFAULT_SESSION_CONF);
+	}
+}
